/*
 * ramchar.c - A simple RAM based character device.
 *
 * (C) Copyright 2004 Jon Masters for Linux User & Developer
 * This program is distributed under the GPL.
 *
 */

#define RAMCHAR_VERSION "1.0"    /* release information */
#define RAMCHAR_NAME "ramchar"   /* identification string */
#define RAMCHAR_MINORS 1         /* number of ramchar devices */

#include <linux/module.h>        /* module macros */
#include <linux/init.h>          /* init macros */

#include <linux/config.h>
#include <linux/errno.h>
#include <linux/kernel.h>

#include <linux/types.h>         /* for custom kernel types */
#include <linux/fs.h>            /* file_operations struct */
#include <linux/fcntl.h>         /* file seeking */
#include <linux/cdev.h>          /* cdev functions */
#include <linux/vmalloc.h>       /* vmalloc functions */
#include <asm/uaccess.h>         /* userspace access */

#define RAMCHAR_SIZE (8192)      /* 8K storage */

static loff_t ramchar_llseek(struct file *file, loff_t offset, int origin);
static ssize_t ramchar_read(struct file *file, char __user *buf,
			    size_t count, loff_t *ppos);
static ssize_t ramchar_write(struct file *file, const char __user *buf,
			     size_t count, loff_t *ppos);
static int ramchar_ioctl(struct inode *inode, struct file *file,
			 unsigned int cmd, unsigned long arg);

struct file_operations ramchar_fops = {
  .owner = THIS_MODULE,
  .llseek = ramchar_llseek,
  .read = ramchar_read,
  .write = ramchar_write,
  .ioctl = ramchar_ioctl,
};

static dev_t ramchar_dev;
static struct cdev *ramchar_cdev;
static u32 *ramchar_mem;

int __init ramchar_init(void)
{

  if (alloc_chrdev_region(&ramchar_dev, 0, RAMCHAR_MINORS, RAMCHAR_NAME)) {
    goto error;
  }

  ramchar_cdev = cdev_alloc();
  kobject_set_name(&ramchar_cdev->kobj,"%s",RAMCHAR_NAME);
  ramchar_cdev->ops = &ramchar_fops;

  if (cdev_add(ramchar_cdev, ramchar_dev, RAMCHAR_MINORS)) {
    kobject_put(&ramchar_cdev->kobj);
    unregister_chrdev_region(ramchar_dev, RAMCHAR_MINORS);
    goto error;
  }

  if (!(ramchar_mem = vmalloc(RAMCHAR_SIZE))) {
    cdev_del(ramchar_cdev);
    unregister_chrdev_region(ramchar_dev, RAMCHAR_MINORS);
    goto error;
  }

  memset(ramchar_mem,'\0',RAMCHAR_SIZE);
  
  printk(KERN_INFO "RAM based character device driver at 0x%08x-0x%08x, version %s\n",
	 (u32)ramchar_mem, (u32)ramchar_mem+RAMCHAR_SIZE, RAMCHAR_VERSION);

  return 0;

 error:
  printk(KERN_ERR "%s: Error registering device.\n",
	 RAMCHAR_NAME);
  return 1;
}

void __exit ramchar_exit(void)
{
  printk(KERN_INFO "%s: Unloading device 0x%08x.\n",
	 RAMCHAR_NAME, (u32)ramchar_mem);
  cdev_del(ramchar_cdev);
  unregister_chrdev_region(ramchar_dev, RAMCHAR_MINORS);
  vfree(ramchar_mem);
}

static loff_t ramchar_llseek(struct file *filp, loff_t off, int whence)
{

  long newpos = 0;

  switch(whence) {
  case 0: /* SEEK_SET */
    newpos = off;
    break;

  case 1: /* SEEK_CUR */
    newpos = filp->f_pos + off;
    break;

  case 2: /* SEEK_END */
    newpos = RAMCHAR_SIZE + off;
    break;

  default: /* Impossible */
    return -EINVAL;
  }

  if (newpos<0) {
    return -EINVAL;
  };

  filp->f_pos = newpos;
  return newpos;

}
static ssize_t ramchar_read(struct file *file, char __user *buf,
			    size_t count, loff_t *f_pos)
{

  if (file->f_pos >= RAMCHAR_SIZE) {
    return 0;
  }

  if (file->f_pos+count >= RAMCHAR_SIZE) {
    count = RAMCHAR_SIZE - file->f_pos;
  }

  /*  printk(KERN_INFO "Going to copy %d bytes (offset %d) from 0x%08x to 0x%08x (device 0x%08x).\n",
   *		  count, (u32)(file->f_pos), (u32)(ramchar_mem+(file->f_pos>>2)), (u32)buf, (u32)ramchar_mem);
   */

  copy_to_user(buf, ramchar_mem+(file->f_pos>>2), count);

  *f_pos += count;

  return count;

}

static ssize_t ramchar_write(struct file *file, const char __user *buf,
			     size_t count, loff_t *f_pos)
{

  if (file->f_pos >= RAMCHAR_SIZE) {
    return 0;
  }

  if (file->f_pos+count >= RAMCHAR_SIZE) {
    count = RAMCHAR_SIZE - file->f_pos;
  }

  /* printk(KERN_INFO "Going to copy %d bytes (offset %d) from 0x%08x to 0x%08x (device 0x%08x).\n",
   *	 count, (u32)(file->f_pos), (u32)buf, (u32)(ramchar_mem+(file->f_pos>>2)),(u32)(ramchar_mem));
   */

  copy_from_user(ramchar_mem+(file->f_pos>>2), buf, count);

  *f_pos += count;

  return count;

}

static int ramchar_ioctl(struct inode *inode, struct file *file,
			 unsigned int cmd, unsigned long arg)
{
  return -EFAULT;
}

module_init(ramchar_init);
module_exit(ramchar_exit);

MODULE_AUTHOR("Jon Masters <jcm@jonmasters.org>");
MODULE_LICENSE("GPL");
